module concrete_vector
  use abstract_vector
  implicit none

  ! TRIVIAL_VECTOR_TYPE provides a simple concrete example of the abstract
  ! type defined by VECTOR_CLASS. Other definitions are possible. The
  ! abstract conjugate gradients routine will work with all of them!

  type, extends(vector_class) :: trivial_vector_type
    real :: elements(100)
  contains
    procedure :: assign      => my_assign
    procedure :: negate      => my_negate
    procedure :: subtract    => my_subtract
    procedure :: add         => my_add
    procedure :: multiply_by => my_multiply_by
    procedure :: saxpy       => my_saxpy
  end type trivial_vector_type

contains

  subroutine my_assign (this,v)
    class(trivial_vector_type), intent(inout) :: this
    class(vector_class),        intent(in)    :: v

    select type (v)
    type is (trivial_vector_type)
      this%elements(:) = v%elements(:)
    end select
  end subroutine my_assign

  subroutine my_negate (this)
    class(trivial_vector_type), intent(inout) :: this

    this%elements(:) = -this%elements(:)
  end subroutine my_negate

  subroutine my_subtract (this,v)
    class(trivial_vector_type), intent(inout) :: this
    class(vector_class),        intent(in)    :: v

    select type (v)
    type is (trivial_vector_type)
      this%elements(:) = this%elements(:) - v%elements(:)
    end select
  end subroutine my_subtract

  subroutine my_add (this,v)
    class(trivial_vector_type), intent(inout) :: this
    class(vector_class),        intent(in)    :: v

    select type (v)
    type is (trivial_vector_type)
      this%elements = this%elements(:) + v%elements(:)
    end select
  end subroutine my_add

  subroutine my_multiply_by (this,s)
    class(trivial_vector_type), intent(inout) :: this
    real, intent(in)                          :: s

    this%elements = s*this%elements(:)
  end subroutine my_multiply_by

  subroutine my_saxpy (this,s,v)
    class(trivial_vector_type), intent(inout) :: this
    real, intent(in)                          :: s
    class(vector_class), intent(in)    :: v

    select type (v)
    type is (trivial_vector_type)
      this%elements = this%elements(:) + s*v%elements(:)
    end select
  end subroutine my_saxpy
end module concrete_vector
