program main

  use abstract_conjugate_gradients
  use concrete_vector
  use concrete_gradient
  use concrete_inner_product
  use concrete_simulator

  implicit none

  ! A demonstration program for the abstract conjugate gradient solver.

  type(trivial_vector_type)        :: x_initial, x_final
  type(trivial_gradient_type)      :: g_initial, g_final
  type(trivial_inner_product_type) :: ipio, ipmin
  type(trivial_simulator_type)     :: sim4d
  integer :: n_iters = 20

  integer :: j

!--- initialise the inner products.
!---   gradients input and output from cg are defined with respect
!---   to the ipio inner product. However, the minimisation is
!---   performed with respect to the ipmin inner product. In a practical
!---   application, this could be used for pre-conditioning.

  ipio%scalp(:)           =  10.0
  ipio%one_over_scalp(:)  =  1.0/ipio%scalp(:)
  ipio%name = 'ipio'

  do j=1,size(ipmin%scalp(:))
    ipmin%scalp(j)          =  j+sqrt(real(j))
    ipmin%one_over_scalp(j) =  1.0/ipmin%scalp(j)
  enddo
  ipmin%name = 'ipmin'

!--- starting point for the minimisation

  x_initial%elements(:) = 7.0

!--- define the initial gradients (including their inner products!)

  g_initial%elements(:) = 0.0
  g_final%elements(:) = 0.0

  call g_initial%change_inner_product_to(ipio)
  call g_final%change_inner_product_to(ipio)

  call sim4d%eval(x_initial,g_initial,ipio)

  print *,'before call cg, initial gradient norm squared=', &
       & ipio%dot_g_g(g_initial,g_initial),' (',ipio%name,' inner product)'

  print *,'calling cg...'

  call cg(x_initial,g_initial,x_final,g_final,n_iters,sim4d,ipmin)

  print *,'after call cg, initial gradient norm squared=', &
       & ipio%dot_g_g(g_initial,g_initial),' (',ipio%name,' inner product)'

  print *,'after call cg, final gradient norm squared=', &
       & ipio%dot_g_g(g_final,g_final),' (',ipio%name,' inner product)'

end program main
