MODULE factory_pattern_2

  TYPE, ABSTRACT :: Connection
  CONTAINS
    PROCEDURE(generic_desc), DEFERRED, PASS(self) :: description
  END TYPE Connection

  ABSTRACT INTERFACE
    SUBROUTINE generic_desc(self)
      IMPORT :: Connection
      CLASS(Connection), INTENT(in) :: self
    END SUBROUTINE generic_desc
  END INTERFACE

  INTERFACE ASSIGNMENT(=)
    module procedure Create_Connection
  END INTERFACE

  TYPE, EXTENDS(Connection) :: PostgresqlConnection
  CONTAINS
    PROCEDURE, PASS(self) :: description => postgresql_desc
  END TYPE PostgresqlConnection
  
  TYPE, EXTENDS(Connection) :: MySQLConnection
  CONTAINS
    PROCEDURE, PASS(self) :: description => mysql_desc
  END TYPE MySQLConnection

CONTAINS
  
  SUBROUTINE postgresql_desc(self)
    CLASS(PostgresqlConnection), INTENT(in) :: self
    WRITE(*,'(A)') "You are now connected with Postgresql"
  END SUBROUTINE postgresql_desc

  SUBROUTINE mysql_desc(self)
    CLASS(MySQLConnection), INTENT(in) :: self
    WRITE(*,'(A)')  "You are now connected with MySQL"
  END SUBROUTINE mysql_desc

  SUBROUTINE create_connection(self, which)  
    CLASS(Connection), ALLOCATABLE, INTENT(OUT) :: self
    CHARACTER(len=*), INTENT(IN) :: which
    
    IF (which == "Postgresql") THEN
      ALLOCATE(PostgresqlConnection :: self)
    ELSE IF (which == "MySQL") THEN
      ALLOCATE(MySQLConnection :: self)
    END IF
  END SUBROUTINE create_connection
END MODULE factory_pattern_2

PROGRAM main
  USE factory_pattern_2
  IMPLICIT NONE
    
  CLASS(Connection), ALLOCATABLE :: db_connect 
  
  db_connect = "Postgresql"
  CALL db_connect%description()
  
  db_connect = "MySQL"
  CALL db_connect%description()
  
END PROGRAM main
